#!/usr/bin/env python

# Copyright 2018 Earth Sciences Department, BSC-CNS
#
# This file is part of HERMESv3_GR.
#
# HERMESv3_GR is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HERMESv3_GR is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HERMESv3_GR. If not, see <http://www.gnu.org/licenses/>.


from hermesv3_gr.modules.writing.writer import Writer
import timeit
from hermesv3_gr.config import settings
import os
import sys
import numpy as np
from netCDF4 import Dataset
from mpi4py import MPI


class WriterCmaq(Writer):

    def __init__(self, path, grid, levels, date, hours, global_attributes_path, compress=True, parallel=False):
        super(WriterCmaq, self).__init__(path, grid, levels, date, hours, global_attributes_path, compress, parallel)

        self.global_attributes_order = [
            'IOAPI_VERSION', 'EXEC_ID', 'FTYPE', 'CDATE', 'CTIME', 'WDATE', 'WTIME', 'SDATE', 'STIME', 'TSTEP', 'NTHIK',
            'NCOLS', 'NROWS', 'NLAYS', 'NVARS', 'GDTYP', 'P_ALP', 'P_BET', 'P_GAM', 'XCENT', 'YCENT', 'XORIG', 'YORIG',
            'XCELL', 'YCELL', 'VGTYP', 'VGTOP', 'VGLVLS', 'GDNAM', 'UPNAM', 'FILEDESC', 'HISTORY', 'VAR-LIST']

    def unit_change(self, variable, data):
        from cf_units import Unit

        if data is not None:
            units = None
            for var in self.variables_attributes:
                if var['name'] == variable:
                    units = var['units']
                    break

            if Unit(units).symbol == Unit('mol.s-1').symbol:
                data = data * 1000 * self.grid.cell_area
            elif Unit(units).symbol == Unit('g.s-1').symbol:
                data = data * 1000 * self.grid.cell_area
            else:
                settings.write_log('ERROR: Check the .err file to get more info.')
                if settings.rank == 0:
                    raise TypeError("The unit '{0}' of specie {1} is not defined correctly. ".format(units, variable) +
                                    "Should be 'mol.s-1.m-2' or 'kg.s-1.m-2'")
                sys.exit(1)
        return data

    @staticmethod
    def change_variable_attributes(emission_list):
        """
        Modifies the emission list to be consistent to use the output as input for CMAQ model.

        :param emission_list: List of emissions
        :type emission_list: list

        :return: Emission list ready for CMAQ
        :rtype: list
        """
        from cf_units import Unit

        var_list = []
        for i in xrange(len(emission_list)):
            emission_list[i]['var_desc'] = "{:<80}".format(emission_list[i]['long_name'])
            emission_list[i]['long_name'] = "{:<16}".format(emission_list[i]['name'])

            if Unit(emission_list[i]['units']).symbol == Unit('mol.s-1').symbol:
                emission_list[i]['units'] = "{:<16}".format('mole/s')
            elif Unit(emission_list[i]['units']).symbol == Unit('g.s-1').symbol:
                emission_list[i]['units'] = "{:<16}".format('g/s')
            var_list.append(emission_list[i]['name'])
        return emission_list, var_list

    @staticmethod
    def create_tflag(st_date, hours_array, num_vars):
        """
        Creates the content of the CMAQ variable TFLAG

        :param st_date: Starting date
        :type st_date: datetime.datetime

        :param hours_array: Array with as elements as time steps. Each element has the delta hours from the starting
                date.
        :type hours_array: numpy.array

        :param num_vars: Number of variables that will contain the NetCDF.
        :type num_vars: int

        :return: Array with the content of TFLAG
        :rtype: numpy.array
        """
        import numpy as np
        from datetime import timedelta

        a = np.array([[[]]])

        for inc_hours in hours_array:
            date = st_date + timedelta(hours=inc_hours)
            b = np.array([[int(date.strftime('%Y%j'))], [int(date.strftime('%H%M%S'))]] * num_vars)
            a = np.append(a, b)

        a.shape = (len(hours_array), 2, num_vars)
        return a

    @staticmethod
    def str_var_list(var_list):
        """
        Transform a list to a string with the elements with 16 white spaces.

        :param var_list: List of variables.
        :type var_list: list

        :return: List transformed on string.
        :rtype: str
        """

        str_var_list = ""
        for var in var_list:
            str_var_list += "{:<16}".format(var)

        return str_var_list

    def read_global_attributes(self):
        import pandas as pd
        from warnings import warn as warning
        float_atts = ['VGTOP']
        int_atts = ['FTYPE', 'NTHIK', 'VGTYP']
        str_atts = ['EXEC_ID', 'GDNAM']
        list_float_atts = ['VGLVLS']

        atts_dict = {
            'EXEC_ID': "{:<80}".format('0.1alpha'),
            'FTYPE': np.int32(1),
            'NTHIK': np.int32(1),
            'VGTYP': np.int32(7),
            'VGTOP': np.float32(5000.),
            'VGLVLS': np.array([1., 0.], dtype=np.float32),
            'GDNAM': "{:<16}".format(''),
        }

        if self.global_attributes_path is not None:
            df = pd.read_csv(self.global_attributes_path)

            for att in atts_dict.iterkeys():
                try:
                    if att in int_atts:
                        atts_dict[att] = np.int32(df.loc[df['attribute'] == att, 'value'].item())
                    elif att in float_atts:
                        atts_dict[att] = np.float32(df.loc[df['attribute'] == att, 'value'].item())
                    elif att in str_atts:
                        atts_dict[att] = str(df.loc[df['attribute'] == att, 'value'].item())
                    elif att in list_float_atts:
                        atts_dict[att] = np.array(df.loc[df['attribute'] == att, 'value'].item().split(),
                                                  dtype=np.float32)
                except ValueError:
                    settings.write_log('WARNING: The global attribute {0} is not defined;'.format(att) +
                                       ' Using default value {0}'.format(atts_dict[att]))
                    if settings.rank == 0:
                        warning('WARNING: The global attribute {0} is not defined; Using default value {1}'.format(
                            att, atts_dict[att]))

        else:
            settings.write_log('WARNING: Check the .err file to get more information.')
            message = 'WARNING: No output attributes defined, check the output_attributes'
            message += ' parameter of the configuration file.\nUsing default values:'
            for key, value in atts_dict.iteritems():
                message += '\n\t{0} = {1}'.format(key, value)
            if settings.rank == 0:
                warning(message)

        return atts_dict

    def create_global_attributes(self, var_list):
        """
        Creates the global attributes and the order that they have to be filled.

        :param date: Starting date.
        :type date: datetime.datetime

        :param nx: Number of elements on the x dimension.
        :type nx: int

        :param ny: Number of elements on the y dimension.
        :type ny: int

        :param nlays: Number of vertical layers.
        :type nlays: int

        :param lat_1: Value of lat 1 of the Lambert Conformal Conic projection.
        :type lat_1: float

        :param lat_2: Value of lat 2 of the Lambert Conformal Conic projection.
        :type lat_2: float

        :param lon_0: Value of lon 0 of the Lambert Conformal Conic projection.
        :type lon_0: float

        :param lat_0: Value of lat 0 of the Lambert Conformal Conic projection.
        :type lat_0: float

        :param x_0: X value of the origin.
        :type x_0: float

        :param y_0: Y value of the origin.
        :type y_0: float

        :param inc_x: Increment (in meters) of the x values.
        :type inc_x: float

        :param inc_y: Increment (in meters) of the y values.
        :type inc_x: float

        :param var_list: List of variables.
        :type var_list: list

        :param exec_id: ID of the execution.
        :type exec_id: str

        :param ftype: File data type = [CUSTOM3:1, GRDDED3:2, BNDARY3:3, IDDATA3:4, PROFIL3:5, or SMATRX3:6]
                    (Default = 1)
        :type ftype: int

        :param tstep: time step, coded HHMMSS according to Models-3 conventions.
        :type tstep: int

        :param nthik: For BNDARY3 files, perimeter thickness (cells), or for SMATRX3 files, number of matrix-columns
                    (unused for other file types)
        :type nthik: int

        :param gdtyp: Map projection type
                LATGRD3=1 (Lat-Lon),
                LAMGRD3=2 (Lambert conformal conic),
                MERGRD3=3 (general tangent Mercator),
                STEGRD3=4 (general tangent stereographic),
                UTMGRD3=5 (UTM, a special case of Mercator),
                POLGRD3=6 (polar secant stereographic),
                EQMGRD3=7 (equatorial secant Mercator), or
                TRMGRD3=8 (transverse secant Mercator)
        :type gdtyp: int

        :param vgtype: Vertical coordinate type
                VGSGPH3=1 (hydrostatic sigma-P),
                VGSGPN3=2 (nonhydrostatic sigma-P),
                VGSIGZ3=3 (sigma-Z),
                VGPRES3=4 (pressure (mb)),
                VGZVAL3=5 (Z (m above sea lvl), or
                VGHVAL3=6 (H (m above ground))
        :type vgtype: int

        :param vgtop: Model-top, for sigma vertical-coordinate types
        :type vgtop: int

        :param vglvls: Array of vertical coordinate level values; level 1 of the grid goes from vertical coordinate
            VGLEVELS[0] to VGLEVELS[1], etc.
        :type vglvls: numpy.array

        :param gdnam: Grid Name
        :type gdnam: str

        :return: Dict of global attributes and a list with the keys ordered.
        :rtype: tuple
        """
        # TODO documentation
        from datetime import datetime

        global_attributes = self.read_global_attributes()

        if len(self.hours) > 1:
            tstep = (self.hours[1] - self.hours[0]) * 10000
        else:
            tstep = 1 * 10000

        now = datetime.now()
        global_attributes['IOAPI_VERSION'] = 'None: made only with NetCDF libraries'
        global_attributes['CDATE'] = np.int32(now.strftime('%Y%j'))
        global_attributes['CTIME'] = np.int32(now.strftime('%H%M%S'))
        global_attributes['WDATE'] = np.int32(now.strftime('%Y%j'))
        global_attributes['WTIME'] = np.int32(now.strftime('%H%M%S'))
        global_attributes['SDATE'] = np.int32(self.date.strftime('%Y%j'))
        global_attributes['STIME'] = np.int32(self.date.strftime('%H%M%S'))
        global_attributes['TSTEP'] = np.int32(tstep)
        global_attributes['NLAYS'] = np.int32(len(self.levels))
        global_attributes['NVARS'] = np.int32(len(var_list))
        global_attributes['UPNAM'] = "{:<16}".format('HERMESv3')
        global_attributes['FILEDESC'] = 'Emissions generated by HERMESv3_GR.'
        global_attributes['HISTORY'] = \
            'Code developed by Barcelona Supercomputing Center (BSC, https://www.bsc.es/).' + \
            'Developer: Carles Tena Medina (carles.tena@bsc.es)' + \
            'Reference: Guevara et al., 2018, GMD., in preparation.'
        global_attributes['VAR-LIST'] = self.str_var_list(var_list)

        if self.grid.grid_type == 'lcc':
            global_attributes['GDTYP'] = np.int32(2)
            global_attributes['NCOLS'] = np.int32(self.grid.nx)
            global_attributes['NROWS'] = np.int32(self.grid.ny)
            global_attributes['P_ALP'] = np.float(self.grid.lat_1)
            global_attributes['P_BET'] = np.float(self.grid.lat_2)
            global_attributes['P_GAM'] = np.float(self.grid.lon_0)
            global_attributes['XCENT'] = np.float(self.grid.lon_0)
            global_attributes['YCENT'] = np.float(self.grid.lat_0)
            global_attributes['XORIG'] = np.float(self.grid.x_0) - np.float(self.grid.inc_x) / 2
            global_attributes['YORIG'] = np.float(self.grid.y_0) - np.float(self.grid.inc_y) / 2
            global_attributes['XCELL'] = np.float(self.grid.inc_x)
            global_attributes['YCELL'] = np.float(self.grid.inc_y)

        return global_attributes

    @staticmethod
    def create_cmaq_netcdf(netcdf_path, center_latitudes, center_longitudes, data_list, levels=None, date=None,
                           hours=None, regular_lat_lon=False, rotated=False, nx=None, ny=None, lat_1=None, lat_2=None,
                           lon_0=None, lat_0=None, x_0=None, y_0=None, inc_x=None, inc_y=None):

        data_list, var_list = WriterCmaq.change_variable_attributes(data_list)

        if settings.writing_serial:
            WriterCmaq.write_serial_netcdf(
                netcdf_path, center_latitudes, center_longitudes, data_list,
                levels=levels, date=date, hours=hours,
                global_attributes=WriterCmaq.create_global_attributes(date, nx, ny, len(levels), lat_1, lat_2, lon_0,
                                                                      lat_0, x_0, y_0, inc_x, inc_y, var_list),
                regular_lat_lon=regular_lat_lon,
                rotated=rotated, )
        else:
            WriterCmaq.write_parallel_netcdf(
                netcdf_path, center_latitudes, center_longitudes, data_list,
                levels=levels, date=date, hours=hours,
                global_attributes=WriterCmaq.create_global_attributes(date, nx, ny, len(levels), lat_1, lat_2, lon_0,
                                                                      lat_0, x_0, y_0, inc_x, inc_y, var_list),
                regular_lat_lon=regular_lat_lon,
                rotated=rotated, )

    @staticmethod
    def write_netcdf(netcdf_path, center_latitudes, center_longitudes, data_list, levels=None, date=None, hours=None,
                     global_attributes=None, regular_lat_lon=False, rotated=False):

        import sys
        from netCDF4 import Dataset

        if regular_lat_lon:
            settings.write_log('ERROR: Check the .err file to get more info.')
            if settings.rank == 0:
                raise TypeError('ERROR: Regular Lat Lon grid not implemented for CMAQ')
            sys.exit(1)

        elif rotated:
            settings.write_log('ERROR: Check the .err file to get more info.')
            if settings.rank == 0:
                raise TypeError('ERROR: Rotated grid not implemented for CMAQ')
            sys.exit(1)

        netcdf = Dataset(netcdf_path, mode='w', format="NETCDF4")

        # ===== Dimensions =====
        netcdf.createDimension('TSTEP', len(hours))
        netcdf.createDimension('DATE-TIME', 2)
        netcdf.createDimension('LAY', len(levels))
        netcdf.createDimension('VAR', len(data_list))
        netcdf.createDimension('ROW', center_latitudes.shape[0])
        netcdf.createDimension('COL', center_longitudes.shape[1])

        # ===== Variables =====
        tflag = netcdf.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME',))
        tflag.setncatts({'units': "{:<16}".format('<YYYYDDD,HHMMSS>'), 'long_name': "{:<16}".format('TFLAG'),
                         'var_desc': "{:<80}".format('Timestep-valid flags:  (1) YYYYDDD or (2) HHMMSS')})
        tflag[:] = WriterCmaq.create_tflag(date, hours, len(data_list))

        # Rest of variables
        for variable in data_list:
            var = netcdf.createVariable(variable['name'], 'f', ('TSTEP', 'LAY', 'ROW', 'COL',), zlib=True)
            var.units = variable['units']
            var.long_name = str(variable['long_name'])
            var.var_desc = str(variable['var_desc'])
            var[:] = variable['data']

        # ===== Global attributes =====
        global_attributes, order = global_attributes
        for attribute in order:
            netcdf.setncattr(attribute, global_attributes[attribute])

        netcdf.close()

    def create_parallel_netcdf(self):
        st_time = timeit.default_timer()
        settings.write_log("\tCreating parallel NetCDF file.", level=2)
        # netcdf = Dataset(netcdf_path, mode='w', format="NETCDF4", parallel=True, comm=settings.comm, info=MPI.Info())
        netcdf = Dataset(self.path, mode='w', format="NETCDF4")

        # ===== Dimensions =====
        settings.write_log("\t\tCreating NetCDF dimensions.", level=2)
        # netcdf.createDimension('TSTEP', len(self.hours))
        netcdf.createDimension('TSTEP', None)
        settings.write_log("\t\t\t'TSTEP' dimension: {0}".format('UNLIMITED ({0})'.format(len(self.hours))), level=3)

        netcdf.createDimension('DATE-TIME', 2)
        settings.write_log("\t\t\t'DATE-TIME' dimension: {0}".format(2), level=3)

        netcdf.createDimension('LAY', len(self.levels))
        settings.write_log("\t\t\t'LAY' dimension: {0}".format(len(self.levels)), level=3)

        netcdf.createDimension('VAR', len(self.variables_attributes))
        settings.write_log("\t\t\t'VAR' dimension: {0}".format(len(self.variables_attributes)), level=3)

        netcdf.createDimension('ROW', self.grid.center_latitudes.shape[0])
        settings.write_log("\t\t\t'ROW' dimension: {0}".format(self.grid.center_latitudes.shape[0]), level=3)

        netcdf.createDimension('COL', self.grid.center_longitudes.shape[1])
        settings.write_log("\t\t\t'COL' dimension: {0}".format(self.grid.center_longitudes.shape[1]), level=3)

        # ===== Variables =====
        settings.write_log("\t\tCreating NetCDF variables.", level=2)
        tflag = netcdf.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME',))
        tflag.setncatts({'units': "{:<16}".format('<YYYYDDD,HHMMSS>'), 'long_name': "{:<16}".format('TFLAG'),
                         'var_desc': "{:<80}".format('Timestep-valid flags:  (1) YYYYDDD or (2) HHMMSS')})
        tflag[:] = self.create_tflag(self.date, self.hours, len(self.variables_attributes))
        settings.write_log("\t\t\t'TFLAG' variable created with size: {0}".format(tflag[:].shape), level=3)

        index = 0
        data_list, var_list = self.change_variable_attributes(self.variables_attributes)
        for variable in self.variables_attributes:
            index += 1
            var = netcdf.createVariable(variable['name'], 'f', ('TSTEP', 'LAY', 'ROW', 'COL',), zlib=self.compress)
            var.units = variable['units']
            var.long_name = str(variable['long_name'])
            var.var_desc = str(variable['var_desc'])
            settings.write_log("\t\t\t'{0}' variable created with size: {1}".format(variable['name'], var[:].shape) +
                               "\n\t\t\t\t'{0}' variable will be filled later.".format(variable['name']), level=3)

        # ===== Global attributes =====
        settings.write_log("\t\tCreating NetCDF metadata.", level=2)
        global_attributes = self.create_global_attributes(var_list)
        for attribute in self.global_attributes_order:
            netcdf.setncattr(attribute, global_attributes[attribute])

        netcdf.close()

        settings.write_time('WriterCmaq', 'create_parallel_netcdf', timeit.default_timer() - st_time, level=3)

    def write_parallel_netcdf(self, emission_list):
        st_time = timeit.default_timer()
        settings.write_log("\tAppending data to parallel NetCDF file.", level=2)

        if settings.size > 1:
            netcdf = Dataset(self.path, mode='a', format="NETCDF4", parallel=True, comm=settings.comm, info=MPI.Info())
        else:
            netcdf = Dataset(self.path, mode='a', format="NETCDF4")
        settings.write_log("\t\tParallel NetCDF file ready to write.", level=2)
        index = 0
        for variable in self.variables_attributes:
            data = self.calculate_data_by_var(variable['name'], emission_list, self.grid.shape)

            index += 1

            var = netcdf.variables[variable['name']]
            if settings.size > 1:
                var.set_collective(True)
            # Correcting NAN
            if data is None:
                data = 0
            var[:, :, self.grid.x_lower_bound:self.grid.x_upper_bound, self.grid.y_lower_bound:self.grid.y_upper_bound] = data
            settings.write_log("\t\t\t'{0}' variable filled".format(variable['name']))

        netcdf.close()
        settings.write_time('WriterCmaq', 'write_parallel_netcdf', timeit.default_timer() - st_time, level=3)

    def write_serial_netcdf(self, emission_list):
        st_time = timeit.default_timer()

        mpi_numpy = False
        mpi_vector = True

        # Gathering the index
        if mpi_numpy or mpi_vector:
            rank_position = np.array([self.grid.x_lower_bound, self.grid.x_upper_bound, self.grid.y_lower_bound, self.grid.y_upper_bound], dtype='i')
            full_position = None
            if settings.rank == 0:
                full_position = np.empty([settings.size, 4], dtype='i')
            settings.comm.Gather(rank_position, full_position, root=0)

        if settings.rank == 0:
            netcdf = Dataset(self.path, mode='w', format="NETCDF4")

            # ===== Dimensions =====
            settings.write_log("\tCreating NetCDF file.", level=2)
            settings.write_log("\t\tCreating NetCDF dimensions.", level=2)
            netcdf.createDimension('TSTEP', len(self.hours))
            settings.write_log("\t\t\t'TSTEP' dimension: {0}".format(len(self.hours)), level=3)
            netcdf.createDimension('DATE-TIME', 2)
            settings.write_log("\t\t\t'DATE-TIME' dimension: {0}".format(2), level=3)
            netcdf.createDimension('LAY', len(self.levels))
            settings.write_log("\t\t\t'LAY' dimension: {0}".format(len(self.levels)), level=3)
            netcdf.createDimension('VAR', len(self.variables_attributes))
            settings.write_log("\t\t\t'VAR' dimension: {0}".format(len(self.variables_attributes)), level=3)
            netcdf.createDimension('ROW', self.grid.center_latitudes.shape[0])
            settings.write_log("\t\t\t'ROW' dimension: {0}".format(self.grid.center_latitudes.shape[0]), level=3)
            netcdf.createDimension('COL', self.grid.center_longitudes.shape[1])
            settings.write_log("\t\t\t'COL' dimension: {0}".format(self.grid.center_longitudes.shape[1]), level=3)

            # ===== Variables =====
            settings.write_log("\t\tCreating NetCDF variables.", level=2)
            tflag = netcdf.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME',))
            tflag.setncatts({'units': "{:<16}".format('<YYYYDDD,HHMMSS>'), 'long_name': "{:<16}".format('TFLAG'),
                             'var_desc': "{:<80}".format('Timestep-valid flags:  (1) YYYYDDD or (2) HHMMSS')})
            tflag[:] = self.create_tflag(self.date, self.hours, len(self.variables_attributes))
            settings.write_log("\t\t\t'TFLAG' variable created with size: {0}".format(tflag[:].shape), level=3)

        full_shape = None
        index = 0
        data_list, var_list = self.change_variable_attributes(self.variables_attributes)
        for variable in data_list:
            if settings.size != 1:
                settings.write_log("\t\t\tGathering {0} data.".format(variable['name']), level=3)
            rank_data = self.calculate_data_by_var(variable['name'], emission_list, self.grid.shape)
            if mpi_numpy or mpi_vector:
                if rank_data is not None:
                    root_shape = settings.comm.bcast(rank_data.shape, root=0)
                    if full_shape is None:
                        full_shape = settings.comm.allgather(rank_data.shape)
                        # print 'Rank {0} full_shape: {1}\n'.format(settings.rank, full_shape)
            if mpi_numpy:
                if settings.size != 1:
                    if settings.rank == 0:
                        recvbuf = np.empty((settings.size,) + rank_data.shape)
                    else:
                        recvbuf = None
                    if root_shape != rank_data.shape:
                        rank_data_aux = np.empty(root_shape)
                        rank_data_aux[:, :, :, :-1] = rank_data
                        rank_data = rank_data_aux
                    # print 'Rank {0} data.shape {1}'.format(settings.rank, rank_data.shape)
                    settings.comm.Gather(rank_data, recvbuf, root=0)
                else:
                    recvbuf = rank_data
            elif mpi_vector:
                if rank_data is not None:
                    counts_i = self.tuple_to_index(full_shape)
                    rank_buff = [rank_data, counts_i[settings.rank]]
                    if settings.rank == 0:
                        displacements = self.calculate_displacements(counts_i)
                        recvdata = np.empty(sum(counts_i), dtype=settings.precision)
                    else:
                        displacements = None
                        recvdata = None
                    if settings.precision == np.float32:
                        recvbuf = [recvdata, counts_i, displacements, MPI.FLOAT]
                    elif settings.precision == np.float64:
                        recvbuf = [recvdata, counts_i, displacements, MPI.DOUBLE]
                    else:
                        settings.write_log('ERROR: Check the .err file to get more info.')
                        if settings.rank == 0:
                            raise TypeError('ERROR: precision {0} unknown'.format(settings.precision))
                        sys.exit(1)

                    settings.comm.Gatherv(rank_buff, recvbuf, root=0)

            else:
                if settings.size != 1:
                    data = settings.comm.gather(rank_data, root=0)
                else:
                    data = rank_data

            if settings.rank == 0:
                if not (mpi_numpy or mpi_vector):
                    if settings.size != 1:
                        try:
                            data = np.concatenate(data, axis=3)
                        except:
                            data = 0
                st_time = timeit.default_timer()
                index += 1

                var = netcdf.createVariable(variable['name'], 'f', ('TSTEP', 'LAY', 'ROW', 'COL',), zlib=self.compress)
                var.units = variable['units']
                var.long_name = str(variable['long_name'])
                var.var_desc = str(variable['var_desc'])
                # var[:] = variable['data']

                if mpi_numpy:
                    data = np.ones(var[:].shape, dtype=settings.precision) * 100
                    for i in xrange(settings.size):
                        try:
                            if i == 0:
                                var[:, :, :, :full_position[i][3]] = recvbuf[i]
                            elif i == settings.size - 1:
                                var[:, :, :, full_position[i][2]:] = recvbuf[i, :, :, :, :-1]
                            else:
                                var[:, :, :, full_position[i][2]:full_position[i][3]] = \
                                    recvbuf[i, :, :, :, : full_shape[i][-1]]
                        except:
                            settings.write_log('ERROR: Check the .err file to get more info.')
                            if settings.rank == 0:
                                raise TypeError("ERROR on i {0} ".format(i) +
                                                "data shape: {0} ".format(data[:, :, :, full_position[i][2]:].shape) +
                                                "recvbuf shape {0}".format(recvbuf[i].shape))
                            sys.exit(1)

                elif mpi_vector:
                    var_time = timeit.default_timer()

                    # data_list = []#np.empty(shape, dtype=np.float64)

                    if rank_data is not None:
                        data = np.empty(var[:].shape, dtype=settings.precision)
                        for i in xrange(settings.size):
                            # print 'Resizeing {0}'.format(i)
                            if not i == settings.size - 1:
                                data[:, :, full_position[i][0]:full_position[i][1], full_position[i][2]:full_position[i][3]] = np.array(recvbuf[0][displacements[i]: displacements[i + 1]]).reshape(full_shape[i])
                            else:
                                data[:, :, full_position[i][0]:full_position[i][1], full_position[i][2]:full_position[i][3]] = np.array(recvbuf[0][displacements[i]:]).reshape(full_shape[i])
                    else:
                        data = 0
                    var[:] = data
                else:
                    var[:] = data
                settings.write_log("\t\t\t'{0}' variable created with size: {1}".format(variable['name'], var[:].shape),
                                   level=3)
        settings.write_log("\t\tCreating NetCDF metadata.", level=2)
        if settings.rank == 0:
            # ===== Global attributes =====
            global_attributes = self.create_global_attributes(var_list)
            for attribute in self.global_attributes_order:
                netcdf.setncattr(attribute, global_attributes[attribute])

            netcdf.close()
        settings.write_time('WriterCmaq', 'write_serial_netcdf', timeit.default_timer() - st_time, level=3)
        return True
